import asyncio
import errno
import grp
import hashlib
import hmac
import io
import json
import logging
import os
import pwd
import re
import signal
import subprocess
import tarfile
import time
from asyncio import Semaphore
from base64 import b64decode
from collections.abc import AsyncIterator, Iterable
from contextlib import asynccontextmanager
from datetime import timedelta
from enum import StrEnum
from functools import cached_property
from importlib.metadata import version
from pathlib import Path
from tempfile import SpooledTemporaryFile, TemporaryDirectory
from typing import TYPE_CHECKING, Any, NamedTuple
from zipfile import BadZipFile

import aioboto3
import httpx
from botocore.config import Config
from pydantic import BaseModel, ConfigDict, RootModel, field_validator

from sagemaker_shim.exceptions import UserSafeError
from sagemaker_shim.extract import safe_extract
from sagemaker_shim.logging import STDOUT_LEVEL

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath  # pragma: no cover
    from types_aiobotocore_s3.client import S3Client  # pragma: no cover
else:
    StrOrBytesPath = object
    S3Client = object


logger = logging.getLogger(__name__)
lock = asyncio.Lock()

BUCKET_NAME_REGEX = re.compile(r"^[a-zA-Z0-9.\-_]{1,255}$")
BUCKET_ARN_REGEX = re.compile(
    r"^arn:(aws).*:(s3|s3-object-lambda):[a-z\-0-9]*:[0-9]{12}:accesspoint[/:][a-zA-Z0-9\-.]{1,63}$|^arn:(aws).*:s3-outposts:[a-z\-0-9]+:[0-9]{12}:outpost[/:][a-zA-Z0-9\-]{1,63}[/:]accesspoint[/:][a-zA-Z0-9\-]{1,63}$"
)


class APIMethod(StrEnum):
    INVOKE = "invoke"
    EXEC = "exec"


class UserInfo(NamedTuple):
    uid: int | None
    gid: int | None
    home: str | None
    groups: list[int]


class ProcUserMixin:
    @property
    def _user(self) -> str:
        user = os.environ.get("GRAND_CHALLENGE_COMPONENT_USER", "")
        logger.info(f"{user=}")
        return user

    @cached_property
    def proc_user(self) -> UserInfo:
        if self._user == "":
            return UserInfo(uid=None, gid=None, home=None, groups=[])

        match = re.fullmatch(
            r"^(?P<user>[A-Za-z][A-Za-z0-9._-]*|[0-9]+)?(:(?P<group>[A-Za-z][A-Za-z0-9._-]*|[0-9]+))?$",
            self._user,
        )

        if match:
            user = match.group("user")
            group = match.group("group")

            logger.info(f"Looking up {user=} {group=}")

            info = self._get_user_info(id_or_name=user)
            group_id = self._get_group_id(id_or_name=group)

            gid = info.gid if group_id is None else group_id

            return UserInfo(
                uid=info.uid,
                gid=gid,
                home=info.home,
                groups=self._put_gid_first(gid=gid, groups=info.groups),
            )
        else:
            raise UserSafeError(
                "Invalid argument for the containers USER instruction"
            )

    @classmethod
    def _get_user_info(cls, id_or_name: str | None) -> UserInfo:
        if id_or_name is None:
            return UserInfo(uid=None, gid=None, home=None, groups=[])

        try:
            user = pwd.getpwnam(id_or_name)
        except (KeyError, AttributeError):
            try:
                uid = int(id_or_name)
            except ValueError as error:
                raise UserSafeError(
                    "The user defined in the containers USER instruction "
                    "does not exist"
                ) from error

            try:
                user = pwd.getpwuid(uid)
            except (KeyError, AttributeError):
                return UserInfo(uid=uid, gid=None, home=None, groups=[])

        return UserInfo(
            uid=user.pw_uid,
            gid=user.pw_gid,
            home=user.pw_dir,
            groups=cls._get_users_groups(user=user),
        )

    @classmethod
    def _get_users_groups(cls, *, user: pwd.struct_passwd) -> list[int]:
        users_groups = [
            g.gr_gid for g in grp.getgrall() if user.pw_name in g.gr_mem
        ]
        return cls._put_gid_first(gid=user.pw_gid, groups=users_groups)

    @staticmethod
    def _put_gid_first(*, gid: int | None, groups: list[int]) -> list[int]:
        if gid is None:
            return groups
        else:
            user_groups = set(groups)

            try:
                user_groups.remove(gid)
            except KeyError:
                pass

            return [gid, *sorted(user_groups)]

    @staticmethod
    def _get_group_id(id_or_name: str | None) -> int | None:
        if id_or_name is None:
            return None

        try:
            return grp.getgrnam(id_or_name).gr_gid
        except (KeyError, AttributeError):
            try:
                return int(id_or_name)
            except ValueError as error:
                raise UserSafeError(
                    "The group defined in the containers USER "
                    "instruction does not exist"
                ) from error


def clean_path(path: Path) -> None:
    if not path.exists():
        return

    for entry in path.iterdir():
        full_path = entry.resolve()

        if entry.is_symlink():
            try:
                entry.chmod(0o700, follow_symlinks=False)
            except NotImplementedError as error:
                if (
                    str(error)
                    == "chmod: follow_symlinks unavailable on this platform"
                ):
                    pass
                else:
                    raise
            entry.unlink()
            continue

        if full_path.is_file():
            full_path.chmod(0o700)
            full_path.unlink()
        elif full_path.is_dir():
            full_path.chmod(0o700)
            clean_path(path=full_path)
            full_path.rmdir()


class S3File(NamedTuple):
    bucket: str
    key: str


class S3Resources(NamedTuple):
    semaphore: Semaphore
    client: S3Client


@asynccontextmanager
async def get_s3_resources() -> AsyncIterator[S3Resources]:
    semaphore = asyncio.Semaphore(
        int(
            os.environ.get("GRAND_CHALLENGE_COMPONENT_ASYNC_CONCURRENCY", "50")
        )
    )
    session = aioboto3.Session()
    boto_config = Config(
        max_pool_connections=int(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_BOTO_MAX_POOL_CONNECTIONS", "120"
            )
        )
    )

    async with session.client(
        "s3",
        endpoint_url=os.environ.get("AWS_S3_ENDPOINT_URL"),
        config=boto_config,
    ) as client:
        yield S3Resources(semaphore=semaphore, client=client)


def parse_s3_uri(*, s3_uri: str) -> S3File:
    pattern = r"^(https|s3)://(?P<bucket>[^/]+)/?(?P<key>.*)$"
    match = re.fullmatch(pattern, s3_uri)

    if not match:
        raise ValueError(f"Not a valid S3 uri, must match pattern {pattern}")

    return S3File(bucket=match.group("bucket"), key=match.group("key"))


async def get_s3_file_content(
    *, s3_uri: str, s3_resources: S3Resources
) -> bytes:
    s3_file = parse_s3_uri(s3_uri=s3_uri)

    content = io.BytesIO()

    async with s3_resources.semaphore:
        await s3_resources.client.download_fileobj(
            Fileobj=content,
            Bucket=s3_file.bucket,
            Key=s3_file.key,
        )

    content.seek(0)

    return content.read()


def validate_bucket_name(v: str) -> str:
    if BUCKET_NAME_REGEX.match(v) or BUCKET_ARN_REGEX.match(v):
        return v
    else:
        raise ValueError("Invalid bucket name")


class ProcUserTarfile(ProcUserMixin, tarfile.TarFile):
    """
    A tarfile that sets the owner of the extracted files to the user and group
    specified in the GRAND_CHALLENGE_COMPONENT_USER environment variable.
    """

    def chown(
        self,
        tarinfo: tarfile.TarInfo,
        targetpath: StrOrBytesPath,
        numeric_owner: bool,
    ) -> None:
        logger.info(f"chown of extracted {targetpath=}")

        if self.proc_user.uid is None and self.proc_user.gid is None:
            # No user or group specified, use the default
            return super().chown(
                tarinfo=tarinfo,
                targetpath=targetpath,
                numeric_owner=numeric_owner,
            )
        else:
            # Do not change owner if the user or group is not set
            uid = -1 if self.proc_user.uid is None else self.proc_user.uid
            gid = -1 if self.proc_user.gid is None else self.proc_user.gid

            logger.debug(f"Changing owner of {targetpath=} to {uid=}, {gid=}")

            os.chown(path=targetpath, uid=uid, gid=gid)


async def download_and_extract_tarball(
    *, s3_uri: str, dest: Path, s3_resources: S3Resources
) -> None:
    s3_file = parse_s3_uri(s3_uri=s3_uri)

    with SpooledTemporaryFile(max_size=4 * 1024 * 1024 * 1024) as f:
        async with s3_resources.semaphore:
            await s3_resources.client.download_fileobj(
                Bucket=s3_file.bucket,
                Key=s3_file.key,
                Fileobj=f,
            )

        f.seek(0)

        try:
            with ProcUserTarfile.open(fileobj=f, mode="r") as tar:
                tar.extractall(path=dest, filter="data")
        except (tarfile.TarError, FileNotFoundError) as error:
            logger.error(error, exc_info=True)
            raise UserSafeError("Tarfile could not be extracted") from error


class AuxiliaryData:
    def __init__(self, *, s3_resources: S3Resources):
        self._s3_resources = s3_resources

    @property
    def model_source(self) -> str | None:
        """s3 URI to a .tar.gz file that is extracted to model_dest"""
        model = os.environ.get("GRAND_CHALLENGE_COMPONENT_MODEL")
        logger.debug(f"{model=}")
        return model

    @property
    def model_dest(self) -> Path:
        model_dest = Path(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_MODEL_DEST", "/opt/ml/model/"
            )
        )
        logger.debug(f"{model_dest=}")
        return model_dest

    @property
    def ground_truth_source(self) -> str | None:
        """s3 URI to a .tar.gz file that is extracted to ground_truth_dest"""
        ground_truth = os.environ.get("GRAND_CHALLENGE_COMPONENT_GROUND_TRUTH")
        logger.debug(f"{ground_truth=}")
        return ground_truth

    @property
    def ground_truth_dest(self) -> Path:
        ground_truth_dest = Path(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_GROUND_TRUTH_DEST",
                "/opt/ml/input/data/ground_truth/",
            )
        )
        logger.debug(f"{ground_truth_dest=}")
        return ground_truth_dest

    @property
    def writable_directories(self) -> list[Path]:
        writable_directories = [
            Path(d)
            for d in os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_WRITABLE_DIRECTORIES",
                "/opt/ml/output/data:/opt/ml/model:/opt/ml/input/data/ground_truth:/opt/ml/checkpoints:/tmp",
            ).split(":")
            if d
        ]
        logger.debug(f"{writable_directories=}")
        return writable_directories

    @property
    def post_clean_directories(self) -> list[Path]:
        post_clean_directories = [
            Path(d)
            for d in os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_POST_CLEAN_DIRECTORIES",
                "/opt/ml/output/data:/opt/ml/model:/opt/ml/input/data/ground_truth",
            ).split(":")
            if d
        ]
        logger.debug(f"{post_clean_directories=}")
        return post_clean_directories

    async def setup(self) -> None:
        logger.info("Setting up Auxiliary Data")

        self.ensure_directories_are_writable()

        try:
            await self.download_model()
        except UserSafeError as error:
            raise UserSafeError(f"Could not setup model: {error}") from error

        try:
            await self.download_ground_truth()
        except UserSafeError as error:
            raise UserSafeError(
                f"Could not setup ground truth: {error}"
            ) from error

    async def teardown(self) -> None:
        logger.info("Cleaning up Auxiliary Data")
        for p in self.post_clean_directories:
            logger.info(f"Cleaning {p=}")
            clean_path(p)

    def ensure_directories_are_writable(self) -> None:
        for directory in self.writable_directories:
            path = Path(directory)
            path.mkdir(exist_ok=True, parents=True)
            path.chmod(mode=0o777)

    async def download_model(self) -> None:
        if self.model_source is not None:
            logger.info(
                f"Downloading model from {self.model_source=} to {self.model_dest=}"
            )
            self.model_dest.mkdir(parents=True, exist_ok=True)
            await download_and_extract_tarball(
                s3_uri=self.model_source,
                dest=self.model_dest,
                s3_resources=self._s3_resources,
            )

    async def download_ground_truth(self) -> None:
        if self.ground_truth_source is not None:
            logger.info(
                f"Downloading ground truth from {self.ground_truth_source=} "
                f"to {self.ground_truth_dest=}"
            )
            self.ground_truth_dest.mkdir(parents=True, exist_ok=True)
            await download_and_extract_tarball(
                s3_uri=self.ground_truth_source,
                dest=self.ground_truth_dest,
                s3_resources=self._s3_resources,
            )


class InferenceIO(BaseModel):
    """A single input or output file for an inference job"""

    model_config = ConfigDict(frozen=True)

    relative_path: Path
    bucket_name: str
    bucket_key: str
    decompress: bool = False

    @field_validator("bucket_name")
    @classmethod
    def validate_bucket_name(cls, v: str) -> str:
        return validate_bucket_name(v)

    def local_file(self, path: Path) -> Path:
        """The local location of the file"""
        return path / self.relative_path

    async def download(
        self, *, input_path: Path, s3_resources: S3Resources
    ) -> None:
        """Download this file from s3"""
        dest_file = self.local_file(path=input_path)

        logger.info(
            f"Downloading {self.bucket_key=} from {self.bucket_name=} to {dest_file=}"
        )

        dest_file.parent.mkdir(exist_ok=True, parents=True)

        if self.decompress:
            with TemporaryDirectory() as tmp_dir:
                zipfile = Path(tmp_dir) / "src.zip"

                with zipfile.open("wb") as f:
                    async with s3_resources.semaphore:
                        await s3_resources.client.download_fileobj(
                            Bucket=self.bucket_name,
                            Key=self.bucket_key,
                            Fileobj=f,
                        )

                try:
                    safe_extract(src=zipfile, dest=dest_file.parent)
                except BadZipFile as error:
                    raise UserSafeError(
                        "Input zip file could not be extracted"
                    ) from error
                except OSError as error:
                    if error.errno == errno.ENOSPC:
                        raise UserSafeError(
                            "Contents of zip file too large"
                        ) from error
                    else:
                        raise error
        else:
            async with s3_resources.semaphore:
                await s3_resources.client.download_file(
                    Bucket=self.bucket_name,
                    Key=self.bucket_key,
                    Filename=dest_file,
                )

    async def upload(
        self, *, output_path: Path, s3_resources: S3Resources
    ) -> None:
        """Upload this file to s3"""
        src_file = str(self.local_file(path=output_path))

        logger.info(
            f"Uploading {src_file=} to {self.bucket_name=} with {self.bucket_key=}"
        )

        async with s3_resources.semaphore:
            await s3_resources.client.upload_file(
                Filename=src_file,
                Bucket=self.bucket_name,
                Key=self.bucket_key,
            )


class InferenceResult(BaseModel):
    model_config = ConfigDict(frozen=True)

    pk: str
    return_code: int
    exec_duration: timedelta | None
    invoke_duration: timedelta | None
    outputs: list[InferenceIO]
    sagemaker_shim_version: str = version("sagemaker-shim")


def log_external(*, level: int, msg: str, task_pk: str | None) -> None:
    """Send a message to the external logger"""
    logger.log(
        level=level, msg=msg, extra={"internal": False, "task_pk": task_pk}
    )


class UserProcess(ProcUserMixin):
    def __init__(self) -> None:
        self._process: asyncio.subprocess.Process | None = None
        self._stdout_task: asyncio.Task[None] | None = None
        self._stderr_task: asyncio.Task[None] | None = None
        self._current_task_pk: str | None = None

    @staticmethod
    def decode_b64j(*, encoded: str | None) -> Any:
        """JSON decode a base64 string"""
        if encoded is None:
            return None
        else:
            return json.loads(
                b64decode(encoded.encode("utf-8")).decode("utf-8")
            )

    @property
    def process(self) -> asyncio.subprocess.Process:
        if self._process is None:
            raise RuntimeError(
                "The process needs to be initialised before use"
            )
        else:
            return self._process

    @property
    def stdout_task(self) -> asyncio.Task[None]:
        if self._stdout_task is None:
            raise RuntimeError(
                "The stdout task needs to be initialised before use"
            )
        else:
            return self._stdout_task

    @property
    def stderr_task(self) -> asyncio.Task[None]:
        if self._stderr_task is None:
            raise RuntimeError(
                "The stderr task needs to be initialised before use"
            )
        else:
            return self._stderr_task

    @property
    def cmd(self) -> Any:
        """The original command for the subprocess"""
        cmd = self.decode_b64j(
            encoded=os.environ.get("GRAND_CHALLENGE_COMPONENT_CMD_B64J")
        )
        logger.debug(f"{cmd=}")
        return cmd

    @property
    def entrypoint(self) -> Any:
        """The original entrypoint for the subprocess"""
        entrypoint = self.decode_b64j(
            encoded=os.environ.get("GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J")
        )
        logger.debug(f"{entrypoint=}")
        return entrypoint

    @property
    def extra_groups(self) -> list[int] | None:
        if (
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "True"
            ).lower()
            == "true"
        ):
            return self.proc_user.groups
        else:
            return None

    @property
    def proc_args(self) -> list[str]:
        """
        Implementation of CMD and ENTRYPOINT parsing from
        https://docs.docker.com/engine/reference/builder/#understand-how-cmd-and-entrypoint-interact
                                    | No ENTRYPOINT              | ENTRYPOINT exec_entry p1_entry | ENTRYPOINT ["exec_entry", "p1_entry"]
        No CMD                      | error, not allowed         | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry
        CMD ["exec_cmd", "p1_cmd"]  | exec_cmd p1_cmd            | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry exec_cmd p1_cmd
        CMD ["p1_cmd", "p2_cmd"]    | p1_cmd p2_cmd              | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry p1_cmd p2_cmd
        CMD exec_cmd p1_cmd         | /bin/sh -c exec_cmd p1_cmd | /bin/sh -c exec_entry p1_entry | exec_entry p1_entry /bin/sh -c exec_cmd p1_cmd
        """  # noqa: B950
        entrypoint = self.entrypoint
        cmd = self.cmd

        if entrypoint is None and cmd is None:
            raise ValueError("Either cmd or entrypoint must be set")
        elif isinstance(entrypoint, str):
            proc_args = ["/bin/sh", "-c", entrypoint]
        else:
            proc_args = entrypoint if entrypoint is not None else []

            if cmd is None:
                pass
            elif isinstance(cmd, str):
                proc_args += ["/bin/sh", "-c", cmd]
            else:
                proc_args += cmd

        return proc_args

    @property
    def proc_env(self) -> dict[str, str]:
        """The environment for the subprocess"""
        # We include some secret values in the environment so
        # ensure that they're not passed through to the subprocess
        env = {
            key: value
            for key, value in os.environ.items()
            if not key.casefold().startswith("grand_challenge_")
        }

        # Set LD_LIBRARY_PATH correctly, see
        # https://pyinstaller.org/en/stable/runtime-information.html#ld-library-path-libpath-considerations
        lp_key = "LD_LIBRARY_PATH"
        lp_orig = env.get(lp_key + "_ORIG")

        if lp_orig is not None:
            env[lp_key] = lp_orig
        else:
            env.pop(lp_key, None)

        if self.proc_user.home is not None:
            env["HOME"] = self.proc_user.home

        return env

    @property
    def start_and_health_timeout(self) -> timedelta:
        seconds = int(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_START_AND_HEALTH_TIMEOUT_SECONDS",
                "300",
            )
        )
        health_check_timeout = timedelta(seconds=seconds)
        logger.debug(f"{health_check_timeout=}")
        return health_check_timeout

    @property
    def health_check_call_timeout(self) -> timedelta:
        seconds = int(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_HEALTH_CHECK_CALL_TIMEOUT_SECONDS",
                "10",
            )
        )
        health_check_call_timeout = timedelta(seconds=seconds)
        logger.debug(f"{health_check_call_timeout=}")
        return health_check_call_timeout

    @property
    def port(self) -> int:
        user_webserver_port = int(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_USER_WEBSERVER_PORT", "4743"
            )
        )
        logger.debug(f"{user_webserver_port=}")
        return user_webserver_port

    @property
    def api_method(self) -> APIMethod:
        api_method = APIMethod(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_API_METHOD", "exec"
            ).lower()
        )
        logger.debug(f"{api_method=}")
        return api_method

    async def setup(self) -> None:
        if self.api_method == APIMethod.EXEC:
            return
        elif self.api_method == APIMethod.INVOKE:
            logger.info("Setting up user process for invoke mode")

            try:
                await asyncio.wait_for(
                    self.start_process_and_wait_for_health(),
                    timeout=self.start_and_health_timeout.total_seconds(),
                )
            except TimeoutError as error:
                await self.teardown()
                raise UserSafeError(
                    "Health check time limit exceeded"
                ) from error
            except UserSafeError as error:
                await self.teardown()
                log_external(
                    level=logging.ERROR,
                    msg=str(error),
                    task_pk=self._current_task_pk,
                )
                raise UserSafeError(
                    "Failed to start the algorithm container's inference server"
                ) from error

            logger.info("User process is healthy")
        else:
            raise NotImplementedError

    async def teardown(self) -> None:
        if self.api_method == APIMethod.EXEC:
            return
        elif self.api_method == APIMethod.INVOKE:
            logger.info("Tearing down user process")

            if self._process is not None:
                await asyncio.shield(self._terminate_group_and_wait())

            if self._stdout_task is not None:
                try:
                    await self.stdout_task
                except Exception as error:
                    logger.warning(
                        f"Error gathering stdout stream task: {error}",
                        exc_info=True,
                    )

            if self._stderr_task is not None:
                try:
                    await self.stderr_task
                except Exception as error:
                    logger.warning(
                        f"Error gathering stderr stream task: {error}",
                        exc_info=True,
                    )

            logger.info("User process teardown complete")
        else:
            raise NotImplementedError

    async def run_inference(self, *, task: "InferenceTask") -> int:
        self._current_task_pk = task.pk
        if self.api_method == APIMethod.EXEC:
            return await self.execute()
        elif self.api_method == APIMethod.INVOKE:
            logger.info(f"Calling invoke endpoint for task {task.pk}")
            return await self.invoke(timeout=task.timeout)
        else:
            raise NotImplementedError

    async def _start_user_process_and_stream_tasks(self) -> None:
        logger.info(f"Calling {self.proc_args=}")

        try:
            self._process = await asyncio.create_subprocess_exec(
                *self.proc_args,
                user=self.proc_user.uid,
                group=self.proc_user.gid,
                extra_groups=self.extra_groups,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=self.proc_env,
                # A new process group must be used for cancellation
                start_new_session=True,
                # The following should always be set to protect this environment
                shell=False,
                close_fds=True,
            )
        except PermissionError as error:
            raise UserSafeError(
                "The user defined in the containers USER instruction "
                "does not have permission to execute the command "
                "defined by the containers ENTRYPOINT and CMD instructions"
            ) from error
        except FileNotFoundError as error:
            raise UserSafeError(
                "The command defined by the containers ENTRYPOINT and "
                "CMD instructions does not exist"
            ) from error

        self._stdout_task = asyncio.create_task(
            self._stream_to_external(
                stream=self.process.stdout, level=STDOUT_LEVEL
            )
        )
        self._stderr_task = asyncio.create_task(
            self._stream_to_external(
                stream=self.process.stderr, level=STDOUT_LEVEL + 10
            )
        )

    async def health_check(self) -> None:
        """Wait for the health endpoint to return 200"""
        try:
            max_retries = max(
                1,
                int(
                    self.start_and_health_timeout.total_seconds()
                    / self.health_check_call_timeout.total_seconds()
                ),
            )
        except ZeroDivisionError:
            max_retries = 1
        async with httpx.AsyncClient() as client:
            for _ in range(max_retries):
                logger.info("Calling health endpoint")
                try:
                    response = await client.get(
                        f"http://127.0.0.1:{self.port}/health",
                        timeout=self.health_check_call_timeout.total_seconds(),
                    )
                except httpx.TimeoutException:
                    continue
                except httpx.ConnectError:
                    await asyncio.sleep(
                        self.health_check_call_timeout.total_seconds()
                    )
                    continue

                logger.info(
                    f"Health endpoint returned status code {response.status_code}"
                )

                if response.status_code == 200:
                    break
                elif response.is_redirect:
                    raise UserSafeError(
                        "Health endpoint returned redirect response"
                    )
                else:
                    await asyncio.sleep(
                        self.health_check_call_timeout.total_seconds()
                    )

    async def start_process_and_wait_for_health(self) -> None:
        try:
            await self._start_user_process_and_stream_tasks()
        except UserSafeError as error:
            raise UserSafeError(f"Could not start server: {error}") from error

        try:
            await self.health_check()
        except UserSafeError as error:
            raise UserSafeError(f"Health check failed: {error}") from error

    @staticmethod
    async def _cancel_tasks(*, tasks: Iterable[asyncio.Task[None]]) -> None:
        for t in tasks:
            if not t.done():
                t.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)

    async def _terminate_group_and_wait(self) -> None:  # noqa:C901
        if self.process.returncode is not None:
            return

        try:
            os.killpg(self.process.pid, signal.SIGTERM)
        except ProcessLookupError:
            return
        except Exception:
            try:
                self.process.terminate()
            except Exception as error:
                logger.info(error, exc_info=True)

        try:
            # Wait for graceful termination
            await asyncio.wait_for(
                self.process.wait(),
                timeout=int(
                    os.environ.get(
                        "GRAND_CHALLENGE_COMPONENT_SIGTERM_GRACE_SECONDS", "5"
                    )
                ),
            )
            logger.info("Process group terminated")
            return
        except TimeoutError:
            logger.warning(
                "Process group did not exit within grace period; "
                "escalating to SIGKILL"
            )
        except Exception as error:
            logger.info(error, exc_info=True)

        try:
            os.killpg(self.process.pid, signal.SIGKILL)
        except ProcessLookupError:
            return
        except Exception:
            try:
                self.process.kill()
            except Exception as error:
                logger.info(error, exc_info=True)

        try:
            await self.process.wait()
            logger.info("Process group killed")
        except Exception as error:
            logger.warning(error, exc_info=True)

    async def _stream_to_external(
        self, *, stream: asyncio.StreamReader | None, level: int
    ) -> None:
        """Send the contents of an io stream to the external logs"""
        if stream is None:
            return

        while True:
            try:
                line = await stream.readline()
            except ValueError:
                log_external(
                    level=logging.WARNING,
                    msg="WARNING: A log line was skipped as it was too long",
                    task_pk=self._current_task_pk,
                )
                continue

            if not line:
                break

            log_external(
                level=level,
                msg=line.replace(b"\x00", b"").decode("utf-8"),
                task_pk=self._current_task_pk,
            )

    async def execute(self) -> int:
        """
        Run the original entrypoint and command in a subprocess

        This needs to be as lean as possible as the method is timed
        """
        await self._start_user_process_and_stream_tasks()

        try:
            return_code = await self.process.wait()
            await asyncio.gather(self.stdout_task, self.stderr_task)
            return return_code

        except asyncio.CancelledError:
            logger.info("Execution was cancelled")
            # shield so termination completes even if cancellation continues
            await asyncio.shield(self._terminate_group_and_wait())
            await self._cancel_tasks(
                tasks=(self.stdout_task, self.stderr_task)
            )
            raise

        except Exception as error:
            logger.error(error, exc_info=True)
            await asyncio.shield(self._terminate_group_and_wait())
            await self._cancel_tasks(
                tasks=(self.stdout_task, self.stderr_task)
            )
            raise

        finally:
            # best-effort final cleanup; shielded to avoid being interrupted
            await asyncio.shield(self._terminate_group_and_wait())

    async def invoke(self, *, timeout: timedelta) -> int:
        """Wait for the invoke endpoint to return 201"""
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(
                    f"http://127.0.0.1:{self.port}/invoke",
                    timeout=timeout.total_seconds(),
                )
            except httpx.TimeoutException as error:
                raise UserSafeError("Invoke time limit exceeded") from error
            except httpx.ConnectError as error:
                raise UserSafeError(
                    "Could not connect to invoke endpoint"
                ) from error
            except httpx.HTTPError as error:
                raise UserSafeError(
                    "HTTP error calling invoke endpoint"
                ) from error

            if response.status_code == 201:
                return 0
            else:
                raise UserSafeError(
                    f"Invoke endpoint returned status {response.status_code}, "
                    f"expected 201"
                )


class InferenceTask(BaseModel):
    model_config = ConfigDict(frozen=True)

    pk: str
    inputs: list[InferenceIO]
    output_bucket_name: str
    output_prefix: str
    timeout: timedelta

    @field_validator("output_prefix")
    @classmethod
    def validate_prefix(cls, v: str) -> str:
        if not v:
            raise ValueError("Prefix cannot be blank")

        if v[-1] != "/":
            v += "/"

        return v

    @field_validator("output_bucket_name")
    @classmethod
    def validate_bucket_name(cls, v: str) -> str:
        return validate_bucket_name(v)

    @property
    def input_path(self) -> Path:
        """Local path where the subprocess is expected to read its input files"""
        input_path = Path(
            os.environ.get("GRAND_CHALLENGE_COMPONENT_INPUT_PATH", "/input")
        )
        logger.debug(f"{input_path=}")
        return input_path

    @property
    def linked_input_path(self) -> Path:
        """Local path where the input files will be placed and linked to"""
        linked_input_parent = Path(
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_LINKED_INPUT_PARENT",
                "/opt/ml/input/data/",
            )
        )
        linked_input_path = linked_input_parent / f"{self.pk}-input"
        logger.debug(f"{linked_input_path=}")
        return linked_input_path

    @property
    def output_path(self) -> Path:
        """Local path where the subprocess is expected to write its files"""
        output_path = Path(
            os.environ.get("GRAND_CHALLENGE_COMPONENT_OUTPUT_PATH", "/output")
        )
        logger.debug(f"{output_path=}")
        return output_path

    async def run_inference(
        self, *, user_process: UserProcess, s3_resources: S3Resources
    ) -> InferenceResult:
        """Run the inference on a single case"""
        logger.info(f"Awaiting lock for {self.pk=}")

        await asyncio.wait_for(lock.acquire(), timeout=1.0)

        try:
            logger.info(f"Running inference for {self.pk=}")

            try:
                inference_result = await self._run_inference(
                    user_process=user_process, s3_resources=s3_resources
                )
            except* UserSafeError as exception_group:
                for exception in exception_group.exceptions:
                    log_external(
                        level=logging.ERROR,
                        msg=str(exception),
                        task_pk=self.pk,
                    )

                inference_result = InferenceResult(
                    pk=self.pk,
                    return_code=1,
                    outputs=[],
                    exec_duration=None,
                    invoke_duration=None,
                )

            await self.upload_inference_result(
                inference_result=inference_result, s3_resources=s3_resources
            )
        finally:
            lock.release()

        logger.info(f"Invocation {self.pk=} complete")

        return inference_result

    async def _run_inference(  # noqa:C901
        self, *, user_process: UserProcess, s3_resources: S3Resources
    ) -> InferenceResult:
        try:
            self.reset_io()

            await self.download_input(s3_resources=s3_resources)

            start = time.monotonic()

            try:
                return_code = await asyncio.wait_for(
                    user_process.run_inference(task=self),
                    timeout=self.timeout.total_seconds(),
                )
            except TimeoutError:
                log_external(
                    level=logging.ERROR,
                    msg="Time limit exceeded",
                    task_pk=self.pk,
                )
                return_code = 1

            duration = time.monotonic() - start

            logger.info(f"{return_code=}")

            if return_code == 0:
                outputs = await self.upload_output(s3_resources=s3_resources)
            else:
                outputs = set()

            return InferenceResult(
                pk=self.pk,
                return_code=return_code,
                outputs=outputs,
                exec_duration=(
                    timedelta(seconds=duration)
                    if user_process.api_method == APIMethod.EXEC
                    else None
                ),
                invoke_duration=(
                    timedelta(seconds=duration)
                    if user_process.api_method == APIMethod.INVOKE
                    else None
                ),
            )
        finally:
            self.reset_io()

    def reset_io(self) -> None:
        """Resets the input and output directories"""
        try:
            clean_path(path=self.input_path)
            clean_path(path=self.output_path)
            self.reset_linked_input()
        except Exception as error:
            logger.error(error, exc_info=True)
            raise UserSafeError(
                "The containers input and output directories could not be reset"
            ) from error

    def reset_linked_input(self) -> None:
        """Resets the symlink from the input to the linked directory"""
        if (
            os.environ.get(
                "GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "True"
            ).lower()
            == "true"
        ):
            logger.info(
                f"Setting up linked input from {self.input_path} "
                f"to {self.linked_input_path}"
            )

            if self.input_path.exists():
                if self.input_path.is_symlink():
                    self.input_path.unlink()
                elif self.input_path.is_dir():
                    self.input_path.rmdir()

            if self.linked_input_path.exists():
                self.linked_input_path.rmdir()

            self.linked_input_path.mkdir(parents=True)
            self.linked_input_path.chmod(0o755)

            self.input_path.symlink_to(
                self.linked_input_path, target_is_directory=True
            )
            self.input_path.chmod(0o755)

    async def download_input(self, *, s3_resources: S3Resources) -> None:
        """Download all the inputs to the input path"""
        async with asyncio.TaskGroup() as task_group:
            for input_file in self.inputs:
                task_group.create_task(
                    input_file.download(
                        input_path=self.input_path, s3_resources=s3_resources
                    )
                )

    async def upload_output(
        self, *, s3_resources: S3Resources
    ) -> set[InferenceIO]:
        """Upload all the outputs from the output path to s3"""
        output_path = self.output_path
        outputs: set[InferenceIO] = set()

        async with asyncio.TaskGroup() as task_group:
            for f in output_path.rglob("**/*"):
                if not f.is_file() and not f.is_symlink():
                    logger.warning(f"Skipping {f=}")
                else:
                    relative_path = f.relative_to(output_path)
                    output = InferenceIO(
                        relative_path=relative_path,
                        bucket_key=f"{self.output_prefix}{relative_path}",
                        bucket_name=self.output_bucket_name,
                    )
                    task_group.create_task(
                        output.upload(
                            output_path=self.output_path,
                            s3_resources=s3_resources,
                        )
                    )
                    outputs.add(output)

        return outputs

    async def upload_inference_result(
        self, *, inference_result: InferenceResult, s3_resources: S3Resources
    ) -> None:
        content = inference_result.model_dump_json().encode("utf-8")
        signature = hmac.new(
            key=bytes.fromhex(
                os.environ.get("GRAND_CHALLENGE_COMPONENT_SIGNING_KEY_HEX", "")
            ),
            msg=content,
            digestmod=hashlib.sha256,
        ).hexdigest()

        bucket_key = (
            f"{self.output_prefix}.sagemaker_shim/inference_result.json"
        )

        logger.info(
            f"Uploading {bucket_key=} in "
            f"{self.output_bucket_name=} with {inference_result=}"
        )

        async with s3_resources.semaphore:
            await s3_resources.client.upload_fileobj(
                Fileobj=io.BytesIO(content),
                Bucket=self.output_bucket_name,
                Key=bucket_key,
                ExtraArgs={
                    "Metadata": {"signature_hmac_sha256": signature},
                },
            )


class InferenceTaskList(RootModel[list[InferenceTask]]):
    pass



import logging.config
from asyncio.streams import _DEFAULT_LIMIT
from copy import deepcopy
from uuid import uuid4

import pytest

from sagemaker_shim.logging import LOGGING_CONFIG
from sagemaker_shim.models import UserProcess
from tests.utils import encode_b64j


def test_container_responds_to_ping(client):
    response = client.get("/ping")

    # SageMaker waits for an HTTP 200 status code and an empty body for
    # a successful ping request before sending an invocations request.
    assert response.status_code == 200
    assert response.content == b""


def test_container_responds_to_execution_parameters(client):
    response = client.get("/execution-parameters")

    assert response.json() == {
        "MaxConcurrentTransforms": 1,
        "BatchStrategy": "SINGLE_RECORD",
        "MaxPayloadInMB": 6,
    }


def test_invocations_endpoint(client, tmp_path, monkeypatch, capsys, local_s3):
    # To receive inference requests, the container must have a web server
    # listening on port 8080 and must accept POST requests to the
    # /invocations endpoint.
    input_path = tmp_path / "input"
    linked_input_parent = tmp_path / "linked-input"

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH", str(input_path.absolute())
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_LINKED_INPUT_PARENT",
        str(linked_input_parent),
    )

    pk = str(uuid4())
    data = {
        "pk": pk,
        "inputs": [],
        "output_bucket_name": local_s3.output_bucket_name,
        "output_prefix": f"test/{pk}",
        "timeout": "PT10S",
    }

    input_path = tmp_path / "input"
    output_path = tmp_path / "output"

    input_path.mkdir()
    output_path.mkdir()

    bigfile = tmp_path / "test.txt"
    with open(bigfile, "w") as f:
        # asyncio has a limit to the size of log lines,
        # Anything above this will not be logged
        f.write("a" * (_DEFAULT_LIMIT + 1))
        f.write("\n")

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J",
        encode_b64j(
            val=[
                "sh",
                "-c",
                f"cat {bigfile} && echo hellostdout && echo hellostderr 1>&2",
            ]
        ),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH",
        str(input_path),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_OUTPUT_PATH",
        str(output_path),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")

    debug_log = deepcopy(LOGGING_CONFIG)
    debug_log["root"]["level"] = "DEBUG"
    logging.config.dictConfig(debug_log)

    response = client.post("/invocations", json=data)

    # The logs need to be interpretable by grand challenge
    captured = capsys.readouterr()
    assert (
        '{"log": "hellostdout", "level": "INFO", "source": "stdout", '
        f'"internal": false, "task": "{pk}"}}\n'
    ) in captured.out
    assert (
        '{"log": "return_code=0", "level": "INFO", "source": "stdout", '
        '"internal": true, "task": null}'
    ) in captured.out
    assert captured.err == (
        '{"log": "WARNING: A log line was skipped as it was too long", '
        '"level": "WARNING", "source": "stderr", "internal": false, '
        f'"task": "{pk}"}}\n'
        '{"log": "hellostderr", "level": "WARNING", "source": "stderr", '
        f'"internal": false, "task": "{pk}"}}\n'
    )

    # To obtain inferences, Amazon SageMaker sends a POST request to the
    # inference container. The POST request body contains data from
    # Amazon S3. Amazon SageMaker passes the request to the container,
    # and returns the inference result from the container, saving the
    # data from the response to Amazon S3.
    response = response.json()
    assert response["return_code"] == 0


@pytest.mark.parametrize(
    "cmd,entrypoint,expected",
    (
        (
            None,
            "exec_entry p1_entry",
            ["/bin/sh", "-c", "exec_entry p1_entry"],
        ),
        (None, ["exec_entry", "p1_entry"], ["exec_entry", "p1_entry"]),
        (["exec_cmd", "p1_cmd"], None, ["exec_cmd", "p1_cmd"]),
        (
            ["exec_cmd", "p1_cmd"],
            "exec_entry p1_entry",
            ["/bin/sh", "-c", "exec_entry p1_entry"],
        ),
        (
            ["exec_cmd", "p1_cmd"],
            ["exec_entry", "p1_entry"],
            ["exec_entry", "p1_entry", "exec_cmd", "p1_cmd"],
        ),
        (["p1_cmd", "p2_cmd"], None, ["p1_cmd", "p2_cmd"]),
        (
            ["p1_cmd", "p2_cmd"],
            "exec_entry p1_entry",
            ["/bin/sh", "-c", "exec_entry p1_entry"],
        ),
        (
            ["p1_cmd", "p2_cmd"],
            ["exec_entry", "p1_entry"],
            ["exec_entry", "p1_entry", "p1_cmd", "p2_cmd"],
        ),
        ("exec_cmd p1_cmd", None, ["/bin/sh", "-c", "exec_cmd p1_cmd"]),
        (
            "exec_cmd p1_cmd",
            "exec_entry p1_entry",
            ["/bin/sh", "-c", "exec_entry p1_entry"],
        ),
        (
            "exec_cmd p1_cmd",
            ["exec_entry", "p1_entry"],
            ["exec_entry", "p1_entry", "/bin/sh", "-c", "exec_cmd p1_cmd"],
        ),
    ),
)
def test_proc_args(cmd, entrypoint, expected, monkeypatch):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J",
        encode_b64j(val=entrypoint),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    p = UserProcess()

    assert p.proc_args == expected


@pytest.mark.parametrize(
    "envvars",
    (
        (),
        ("GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J",),
        (
            "GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J",
            "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        ),
        ("GRAND_CHALLENGE_COMPONENT_CMD_B64J"),
    ),
)
def test_unset_cmd_and_entrypoint(envvars, monkeypatch):
    for var in envvars:
        monkeypatch.setenv(var, encode_b64j(val=None))

    p = UserProcess()

    with pytest.raises(ValueError) as e:
        p.proc_args

    assert "Either cmd or entrypoint must be set" in str(e)


@pytest.mark.parametrize(
    "val",
    (
        (None),
        (["exec_cmd", "p1_cmd"]),
        ("exec_cmd p1_cmd"),
        ("c\xf7>"),
        ("👍"),
        ("null"),
    ),
)
def test_decode_b64j(val):
    encoded = encode_b64j(val=val)
    assert UserProcess.decode_b64j(encoded=encoded) == val


def test_decode_returns_none():
    assert UserProcess.decode_b64j(encoded=None) is None



import asyncio
import getpass
import grp
import io
import logging.config
import os
import pwd
import re
import tarfile
from contextlib import nullcontext
from datetime import timedelta
from pathlib import Path
from uuid import uuid4

import httpx
import pytest

from sagemaker_shim.exceptions import UserSafeError
from sagemaker_shim.logging import LOGGING_CONFIG
from sagemaker_shim.models import (
    APIMethod,
    AuxiliaryData,
    InferenceTask,
    ProcUserMixin,
    ProcUserTarfile,
    UserProcess,
    clean_path,
    get_s3_resources,
    validate_bucket_name,
)
from tests.utils import encode_b64j


@pytest.fixture
def algorithm_model():
    model_f = io.BytesIO()
    with tarfile.open(fileobj=model_f, mode="w:gz") as tar:
        content = b"Hello, World!"
        file_info = tarfile.TarInfo("model-file1.txt")
        file_info.size = len(content)
        tar.addfile(file_info, io.BytesIO(content))

        file_info = tarfile.TarInfo("model-sub/model-file2.txt")
        file_info.size = len(content)
        tar.addfile(file_info, io.BytesIO(content))

    model_f.seek(0)

    return model_f


def test_invalid_bucket_name():
    with pytest.raises(ValueError):
        validate_bucket_name("$!#")


def test_blank_prefix():
    with pytest.raises(ValueError) as error:
        InferenceTask(
            pk="test",
            inputs=[],
            output_bucket_name="test",
            output_prefix="",
            timeout=timedelta(),
        )

    assert str(error).startswith(
        "<ExceptionInfo 1 validation error for InferenceTask\n"
        "output_prefix\n  Value error, Prefix cannot be blank "
        "[type=value_error, input_value='', input_type=str]\n"
    )


def test_prefix_slash_appended():
    t = InferenceTask(
        pk="test",
        inputs=[],
        output_bucket_name="test",
        output_prefix="test",
        timeout=timedelta(),
    )
    assert t.output_prefix == "test/"


def test_patching_ld_library_path(monkeypatch):
    """Subprocess must run with the original LD_LIBRARY_PATH set"""
    monkeypatch.setenv("LD_LIBRARY_PATH_ORIG", "special")

    p = UserProcess()

    env = os.environ.copy()

    assert env["LD_LIBRARY_PATH_ORIG"] == "special"
    assert "LD_LIBRARY_PATH" not in env
    assert p.proc_env["LD_LIBRARY_PATH"] == "special"


def test_removing_ld_library_path(monkeypatch):
    """Subprocess must have LD_LIBRARY_PATH removed if set"""
    monkeypatch.setenv("LD_LIBRARY_PATH", "present")

    p = UserProcess()

    env = os.environ.copy()

    assert "LD_LIBRARY_PATH" not in p.proc_env
    assert env["LD_LIBRARY_PATH"] == "present"


def test_all_grand_challenge_env_vars_removed(monkeypatch):
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SIGNING_KEY_HEX", "somekey")
    monkeypatch.setenv("grand_challenge_foo", "bar")

    p = UserProcess()

    env = os.environ.copy()

    assert env["GRAND_CHALLENGE_COMPONENT_SIGNING_KEY_HEX"] == "somekey"
    assert env["grand_challenge_foo"] == "bar"
    assert "GRAND_CHALLENGE_COMPONENT_SIGNING_KEY_HEX" not in p.proc_env
    assert "grand_challenge_foo" not in p.proc_env


ROOT_HOME = pwd.getpwnam("root").pw_dir
ROOT_GROUPS = ProcUserMixin._get_users_groups(user=pwd.getpwnam("root"))
USER_HOME = os.path.expanduser("~")
USER_GROUPS = ProcUserMixin._get_users_groups(
    user=pwd.getpwnam(getpass.getuser())
)


@pytest.mark.parametrize(
    "user,expected_user,expected_group,expected_home,expected_extra_groups",
    (
        ("0", 0, 0, ROOT_HOME, ROOT_GROUPS),
        ("0:0", 0, 0, ROOT_HOME, ROOT_GROUPS),
        (":0", None, 0, None, [0]),
        ("", None, None, None, []),
        ("root", 0, 0, ROOT_HOME, ROOT_GROUPS),
        (f"root:{grp.getgrgid(0).gr_name}", 0, 0, ROOT_HOME, ROOT_GROUPS),
        (f":{grp.getgrgid(0).gr_name}", None, 0, None, [0]),
        ("root:0", 0, 0, ROOT_HOME, ROOT_GROUPS),
        (f"0:{grp.getgrgid(0).gr_name}", 0, 0, ROOT_HOME, ROOT_GROUPS),
        (f":{os.getgid()}", None, os.getgid(), None, [os.getgid()]),
        (
            f"root:{os.getgid()}",
            0,
            os.getgid(),
            ROOT_HOME,
            ProcUserMixin._put_gid_first(gid=os.getgid(), groups=ROOT_GROUPS),
        ),
        # User exists
        (f"{os.getuid()}", os.getuid(), os.getgid(), USER_HOME, USER_GROUPS),
        (
            f"{getpass.getuser()}",
            os.getuid(),
            os.getgid(),
            USER_HOME,
            USER_GROUPS,
        ),
        # Group does not exist, but is an int
        (
            f"{os.getuid()}:23746",
            os.getuid(),
            23746,
            USER_HOME,
            ProcUserMixin._put_gid_first(gid=23746, groups=USER_GROUPS),
        ),
        (
            f"{getpass.getuser()}:23746",
            os.getuid(),
            23746,
            USER_HOME,
            ProcUserMixin._put_gid_first(gid=23746, groups=USER_GROUPS),
        ),
        # User does not exist, but is an int
        ("23746", 23746, None, None, []),
        (f"23746:{grp.getgrgid(0).gr_name}", 23746, 0, None, [0]),
        (f"23746:{os.getgid()}", 23746, os.getgid(), None, [os.getgid()]),
        # User and group do not exist, but are ints
        ("23746:23746", 23746, 23746, None, [23746]),
    ),
)
def test_proc_user(
    monkeypatch,
    user,
    expected_user,
    expected_group,
    expected_home,
    expected_extra_groups,
    algorithm_model,
):
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USER", user)

    p = UserProcess()

    assert p._user == user
    assert p.proc_user.uid == expected_user
    assert p.proc_user.gid == expected_group
    assert p.proc_user.home == expected_home
    assert p.extra_groups == expected_extra_groups
    assert None not in p.extra_groups

    with ProcUserTarfile.open(fileobj=algorithm_model, mode="r") as tar:
        assert tar.proc_user == p.proc_user


def test_put_gid_first():
    # Setting None leaves the groups unmodified
    assert ProcUserMixin._put_gid_first(gid=None, groups=[2, 1, 3, 4]) == [
        2,
        1,
        3,
        4,
    ]
    # Setting an existing group puts it first and orders the rest
    assert ProcUserMixin._put_gid_first(gid=3, groups=[2, 1, 3, 4]) == [
        3,
        1,
        2,
        4,
    ]
    # Adding a group puts it first and orders the rest
    assert ProcUserMixin._put_gid_first(gid=5, groups=[2, 1, 3, 4]) == [
        5,
        1,
        2,
        3,
        4,
    ]


# Should error
@pytest.mark.parametrize(
    "user,expected_error",
    (
        (
            f"{os.getuid()}:nonExistentGroup",
            "The group defined in the containers USER instruction does not exist",
        ),
        (
            f"{getpass.getuser()}:nonExistentGroup",
            "The group defined in the containers USER instruction does not exist",
        ),
        (
            "nonExistentUser",
            "The user defined in the containers USER instruction does not exist",
        ),
        (
            "nonExistentUser:nonExistentGroup",
            "The user defined in the containers USER instruction does not exist",
        ),
        (
            ":nonExistentGroup",
            "The group defined in the containers USER instruction does not exist",
        ),
        (
            f"nonExistentUser:{grp.getgrgid(0).gr_name}",
            "The user defined in the containers USER instruction does not exist",
        ),
        (
            f"nonExistentUser:{os.getgid()}",
            "The user defined in the containers USER instruction does not exist",
        ),
        ("🙈:🙉", "Invalid argument for the containers USER instruction"),
        ("🙈", "Invalid argument for the containers USER instruction"),
        (":🙉", "Invalid argument for the containers USER instruction"),
    ),
)
def test_proc_user_errors(monkeypatch, user, expected_error):
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USER", user)

    p = UserProcess()

    assert p._user == user

    with pytest.raises(UserSafeError) as error:
        _ = p.proc_user

    assert str(error.value) == expected_error


def test_proc_user_unset(algorithm_model):
    p = UserProcess()

    assert p._user == ""
    assert p.proc_user.uid is None
    assert p.proc_user.gid is None
    assert p.proc_user.home is None

    with ProcUserTarfile.open(fileobj=algorithm_model, mode="r") as tar:
        assert tar.proc_user == p.proc_user


def test_home_is_set(monkeypatch):
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USER", "root")

    p = UserProcess()

    assert p.proc_env["HOME"] == pwd.getpwnam("root").pw_dir


@pytest.mark.asyncio
async def test_model_and_ground_truth_extraction(
    local_s3, monkeypatch, tmp_path, mocker, algorithm_model
):
    model_pk = str(uuid4())
    ground_truth_pk = str(uuid4())

    ground_truth_f = io.BytesIO()
    with tarfile.open(fileobj=ground_truth_f, mode="w:gz") as tar:
        content = b"Hello, World!"
        file_info = tarfile.TarInfo("gt-file1.txt")
        file_info.size = len(content)
        tar.addfile(file_info, io.BytesIO(content))

        file_info = tarfile.TarInfo("gt-sub/gt-file2.txt")
        file_info.size = len(content)
        tar.addfile(file_info, io.BytesIO(content))

    ground_truth_f.seek(0)

    model_destination = tmp_path / "model"
    ground_truth_destination = tmp_path / "ground_truth"

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_MODEL",
        f"s3://{local_s3.input_bucket_name}/{model_pk}/model.tar.gz",
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_MODEL_DEST", str(model_destination)
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_GROUND_TRUTH",
        f"s3://{local_s3.input_bucket_name}/{ground_truth_pk}/ground_truth.tar.gz",
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_GROUND_TRUTH_DEST",
        str(ground_truth_destination),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_POST_CLEAN_DIRECTORIES",
        f"{model_destination}:{ground_truth_destination}",
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_WRITABLE_DIRECTORIES", "")

    spy = mocker.spy(ProcUserTarfile, "chown")

    async with get_s3_resources() as s3_resources:
        async with s3_resources.semaphore:
            await s3_resources.client.upload_fileobj(
                algorithm_model,
                local_s3.input_bucket_name,
                f"{model_pk}/model.tar.gz",
            )
            await s3_resources.client.upload_fileobj(
                ground_truth_f,
                local_s3.input_bucket_name,
                f"{ground_truth_pk}/ground_truth.tar.gz",
            )

        auxiliary_data = AuxiliaryData(s3_resources=s3_resources)
        await auxiliary_data.setup()
        downloaded_files = {
            str(f.relative_to(tmp_path))
            for f in tmp_path.rglob("**/*")
            if f.is_file()
        }
        await auxiliary_data.teardown()

    assert downloaded_files == {
        "model/model-file1.txt",
        "model/model-sub/model-file2.txt",
        "ground_truth/gt-file1.txt",
        "ground_truth/gt-sub/gt-file2.txt",
    }

    # Files should be cleaned up
    assert {str(f.relative_to(tmp_path)) for f in tmp_path.rglob("**/*")} == {
        "model",
        "ground_truth",
    }

    # We cannot test chown as you need to be root, but we can test that it was called
    assert spy.call_count == 4


@pytest.mark.asyncio
async def test_ensure_directories_are_writable_unset(monkeypatch):
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_WRITABLE_DIRECTORIES", "")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_POST_CLEAN_DIRECTORIES", "")

    async with get_s3_resources() as s3_resources:
        auxiliary_data = AuxiliaryData(s3_resources=s3_resources)

        assert auxiliary_data.writable_directories == []
        assert auxiliary_data.post_clean_directories == []
        assert auxiliary_data.model_source is None
        assert auxiliary_data.model_dest == Path("/opt/ml/model")
        assert not auxiliary_data.model_dest.exists()
        assert auxiliary_data.ground_truth_source is None
        assert auxiliary_data.ground_truth_dest == Path(
            "/opt/ml/input/data/ground_truth"
        )
        assert not auxiliary_data.ground_truth_dest.exists()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "directories,expected",
    (
        ("", []),
        (":", []),
        ("::", []),
    ),
)
async def test_ensure_directories_are_writable_set(
    monkeypatch, directories, expected
):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_WRITABLE_DIRECTORIES",
        directories,
    )

    async with get_s3_resources() as s3_resources:
        auxiliary_data = AuxiliaryData(s3_resources=s3_resources)
        assert auxiliary_data.writable_directories == expected


@pytest.mark.asyncio
async def test_ensure_directories_are_writable(tmp_path, monkeypatch):
    data = tmp_path / "opt" / "ml" / "output" / "data"
    data.mkdir(mode=0o755, parents=True)

    model = tmp_path / "opt" / "ml" / "model"
    model.mkdir(mode=0o755, parents=True)

    # Do not create the checkpoints dir in the test
    checkpoints = tmp_path / "opt" / "ml" / "checkpoints"

    tmp = tmp_path / "tmp"
    tmp.mkdir(mode=0o755)

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_WRITABLE_DIRECTORIES",
        f"{data.absolute()}:{model.absolute()}:{checkpoints.absolute()}:{tmp.absolute()}",
    )

    async with get_s3_resources() as s3_resources:
        auxiliary_data = AuxiliaryData(s3_resources=s3_resources)
        await auxiliary_data.setup()
        await auxiliary_data.teardown()

    assert data.stat().st_mode == 0o40777
    assert model.stat().st_mode == 0o40777
    assert checkpoints.stat().st_mode == 0o40777
    assert tmp.stat().st_mode == 0o40777


def test_linked_input_path_default():
    t = InferenceTask(
        pk="test",
        inputs=[],
        output_bucket_name="test",
        output_prefix="test",
        timeout=timedelta(),
    )

    assert t.linked_input_path == Path("/opt/ml/input/data/test-input")


def test_linked_input_path_setting(monkeypatch):
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_LINKED_INPUT_PARENT", "/foo")

    t = InferenceTask(
        pk="test",
        inputs=[],
        output_bucket_name="test",
        output_prefix="test",
        timeout=timedelta(),
    )

    assert t.linked_input_path == Path("/foo/test-input")


def test_reset_linked_input(tmp_path, monkeypatch):
    input_path = tmp_path / "input"
    linked_input_parent = tmp_path / "linked-input"

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_INPUT_PATH", str(input_path.absolute())
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_LINKED_INPUT_PARENT",
        str(linked_input_parent),
    )

    t = InferenceTask(
        pk="test",
        inputs=[],
        output_bucket_name="test",
        output_prefix="test",
        timeout=timedelta(),
    )
    t.reset_io()

    expected_input_directory = linked_input_parent / "test-input"

    assert input_path.exists()
    assert input_path.is_symlink()
    assert expected_input_directory.exists()
    assert expected_input_directory.is_dir()
    assert input_path.resolve(strict=True) == expected_input_directory

    # Ensure 0o755 permissions
    assert os.stat(input_path).st_mode & 0o777 == 0o755
    assert os.stat(expected_input_directory).st_mode & 0o777 == 0o755


@pytest.mark.asyncio
async def test_exec_timeout(local_s3, monkeypatch, capsys):
    cmd = ["sleep", "10"]
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    process = UserProcess()
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=str(prefix),
        timeout=timedelta(seconds=1),
    )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 1
    assert int(result.exec_duration.total_seconds()) == 1
    assert result.invoke_duration is None  # should only be set for invocation

    captured = capsys.readouterr()
    # "Time limit exceeded" must be the last log for the user error
    assert captured.err == (
        '{"log": "Time limit exceeded", "level": "ERROR", "source": "stderr", '
        f'"internal": false, "task": "{pk}"}}\n'
    )
    assert (
        '{"log": "Execution was cancelled", "level": "INFO", "source": "stdout", '
        '"internal": true, "task": null}' in captured.out
    )
    assert (
        '{"log": "Process group terminated", "level": "INFO", "source": "stdout", '
        '"internal": true, "task": null}' in captured.out
    )


@pytest.mark.asyncio
async def test_non_existent_user(local_s3, monkeypatch, capsys):
    cmd = ["sleep", "10"]
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    process = UserProcess()
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=str(prefix),
        timeout=timedelta(seconds=1),
    )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USER", "-gdfs")

    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 1
    assert result.exec_duration is None
    assert result.invoke_duration is None  # should only be set for invocation

    captured = capsys.readouterr()
    # Invalid argument must be the last log for the user error
    assert captured.err == (
        '{"log": "Invalid argument for the containers USER instruction", '
        f'"level": "ERROR", "source": "stderr", "internal": false, "task": "{pk}"}}\n'
    )


@pytest.mark.asyncio
async def test_user_cmd_permission_denied(
    local_s3, monkeypatch, capsys, tmp_path
):
    test_file = tmp_path / "no_perms.sh"
    test_file.touch()

    cmd = [str(tmp_path / "no_perms.sh")]
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    process = UserProcess()
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=str(prefix),
        timeout=timedelta(seconds=1),
    )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 1
    assert result.exec_duration is None
    assert result.invoke_duration is None  # should only be set for invocation

    captured = capsys.readouterr()
    # No permission must be the last log for the user error
    assert captured.err == (
        '{"log": "The user defined in the containers USER instruction '
        "does not have permission to execute the command defined by "
        'the containers ENTRYPOINT and CMD instructions", "level": "ERROR", '
        f'"source": "stderr", "internal": false, "task": "{pk}"}}\n'
    )


@pytest.mark.asyncio
async def test_user_cmd_missing(local_s3, monkeypatch, capsys):
    cmd = ["doesnt_exist.sh"]
    pk = str(uuid4())
    prefix = f"tasks/{pk}"
    process = UserProcess()
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=str(prefix),
        timeout=timedelta(seconds=1),
    )

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=cmd),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")

    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 1
    assert result.exec_duration is None
    assert result.invoke_duration is None  # should only be set for invocation

    captured = capsys.readouterr()
    # Command not found must be the last log for the user error
    assert captured.err == (
        '{"log": "The command defined by the containers ENTRYPOINT '
        'and CMD instructions does not exist", "level": "ERROR", '
        f'"source": "stderr", "internal": false, "task": "{pk}"}}\n'
    )


def test_folder_cleanup_with_deleted_symlink_targets(tmp_path):
    dirty_path = tmp_path / "test"
    dirty_path.mkdir()

    # Create a target directory
    target_dir = tmp_path / "target-dir"
    target_dir.mkdir()

    # Create a symlink to the target directory
    dir_symlink = dirty_path / "to-target-dir"
    dir_symlink.symlink_to(target=target_dir, target_is_directory=True)

    # Create a target file
    target_file = tmp_path / "target-file"
    target_file.touch()

    # Create a symlink to the target file
    file_symlink = dirty_path / "to-target-file"
    file_symlink.symlink_to(target=target_file, target_is_directory=False)

    # Remove the targets so the symlinks do not point at anything
    target_dir.rmdir()
    target_file.unlink()

    clean_path(dirty_path)

    assert {*dirty_path.iterdir()} == set()
    assert dirty_path.exists()

    assert not target_dir.exists()
    assert not dir_symlink.exists()
    assert not target_file.exists()
    assert not file_symlink.exists()


def test_symlink_cleanup(tmp_path):
    dirty_path = tmp_path / "test"
    dirty_path.mkdir()

    # Create a target directory
    target_dir = tmp_path / "target-dir"
    target_dir.mkdir()

    # Create a symlink to the target directory
    dir_symlink = dirty_path / "to-target-dir"
    dir_symlink.symlink_to(target=target_dir, target_is_directory=True)

    # Create a target file
    target_file = tmp_path / "target-file"
    target_file.touch()

    # Create a symlink to the target file
    file_symlink = dirty_path / "to-target-file"
    file_symlink.symlink_to(target=target_file, target_is_directory=False)

    clean_path(dirty_path)

    assert {*dirty_path.iterdir()} == set()
    assert dirty_path.exists()

    assert target_dir.exists()
    assert not dir_symlink.exists()
    assert target_file.exists()
    assert not file_symlink.exists()


def test_circular_cleanup(tmp_path):
    dirty_path = tmp_path / "clean-me"
    dirty_path.mkdir()

    dirty_path_dir = dirty_path / "dirty-dir"
    dirty_path_dir.mkdir()

    dirty_path_file = dirty_path / "dirty-file"
    dirty_path_file.touch()

    # Create a target directory
    target_dir = tmp_path / "target-dir"
    target_dir.mkdir()

    target_dir_symlink_back = target_dir / "dirty-dir"
    target_dir_symlink_back.symlink_to(
        dirty_path_dir, target_is_directory=True
    )

    target_dir_symlink_file_back = target_dir / "dirty-file"
    target_dir_symlink_file_back.symlink_to(
        dirty_path_file, target_is_directory=False
    )

    # Create a symlink to the target directory
    dir_symlink = dirty_path / "to-target-dir"
    dir_symlink.symlink_to(target=target_dir, target_is_directory=True)

    # Create a target file
    target_file = tmp_path / "target-file"
    target_file.touch()

    # Create a symlink to the target file
    file_symlink = dirty_path / "to-target-file"
    file_symlink.symlink_to(target=target_file, target_is_directory=False)

    clean_path(dirty_path)

    assert {*dirty_path.iterdir()} == set()
    assert dirty_path.exists()

    assert target_dir.exists()
    assert not dir_symlink.exists()
    assert target_file.exists()
    assert not file_symlink.exists()
    assert not dirty_path_dir.exists()
    assert target_dir_symlink_back.exists(follow_symlinks=False)
    assert target_dir_symlink_file_back.exists(follow_symlinks=False)


def test_user_process_start_and_health_timeout(monkeypatch):
    process = UserProcess()

    assert process.start_and_health_timeout.total_seconds() == 300

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_START_AND_HEALTH_TIMEOUT_SECONDS", "42"
    )

    assert process.start_and_health_timeout.total_seconds() == 42


def test_user_process_health_check_call_timeout(monkeypatch):
    process = UserProcess()

    assert process.health_check_call_timeout.total_seconds() == 10

    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_HEALTH_CHECK_CALL_TIMEOUT_SECONDS", "42"
    )

    assert process.health_check_call_timeout.total_seconds() == 42


def test_user_process_port(monkeypatch):
    process = UserProcess()

    assert process.port == 4743

    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USER_WEBSERVER_PORT", "1442")

    assert process.port == 1442


def test_user_process_api_method(monkeypatch):
    process = UserProcess()

    assert process.api_method == APIMethod.EXEC

    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_API_METHOD", "invoke")

    assert process.api_method == APIMethod.INVOKE


@pytest.mark.parametrize(
    "status_code, expectation",
    [
        (200, nullcontext()),
        (
            300,
            pytest.raises(
                UserSafeError,
                match="Health endpoint returned redirect response",
            ),
        ),
    ],
)
@pytest.mark.asyncio
async def test_health_check(mocker, status_code, expectation):
    mocker.patch(
        "sagemaker_shim.models.httpx.AsyncClient.get",
        return_value=httpx.Response(status_code),
    )

    p = UserProcess()

    with expectation:
        await p.health_check()


@pytest.mark.asyncio
async def test_invoke_start_and_health_timeout(monkeypatch, capsys):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "invoke",
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_START_AND_HEALTH_TIMEOUT_SECONDS",
        "0",
    )
    logging.config.dictConfig(LOGGING_CONFIG)
    p = UserProcess()

    with pytest.raises(
        UserSafeError, match="Health check time limit exceeded"
    ):
        await p.setup()

    captured = capsys.readouterr()
    assert "Setting up user process for invoke mode" in captured.out
    assert "Tearing down user process" in captured.out


@pytest.mark.parametrize(
    "patch_kwargs",
    [
        {"return_value": httpx.Response(100)},
        {"side_effect": httpx.ConnectError("error")},
    ],
)
@pytest.mark.asyncio
async def test_invoke_start_and_health_call_timeout_sleep(
    mocker, monkeypatch, patch_kwargs, capsys
):
    mocker.patch(
        "sagemaker_shim.models.httpx.AsyncClient.get",
        **patch_kwargs,
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "invoke",
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_START_AND_HEALTH_TIMEOUT_SECONDS",
        "2",
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_HEALTH_CHECK_CALL_TIMEOUT_SECONDS",
        "1",
    )
    logging.config.dictConfig(LOGGING_CONFIG)
    p = UserProcess()

    with pytest.raises(
        UserSafeError, match="Health check time limit exceeded"
    ):
        await p.setup()

    captured = capsys.readouterr()
    assert len(re.findall("Calling health endpoint", captured.out)) == 2


@pytest.mark.asyncio
async def test_invoke_setup_and_teardown(mocker, monkeypatch, capsys):
    mocker.patch(
        "sagemaker_shim.models.httpx.AsyncClient.get",
        return_value=httpx.Response(200),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "invoke",
    )
    logging.config.dictConfig(LOGGING_CONFIG)
    p = UserProcess()

    await p.setup()
    await p.teardown()

    captured = capsys.readouterr()
    assert "Setting up user process for invoke mode" in captured.out
    assert "Calling health endpoint" in captured.out
    assert "Health endpoint returned status code 200" in captured.out
    assert "User process is healthy" in captured.out
    assert "Tearing down user process" in captured.out
    assert "User process teardown complete" in captured.out


@pytest.mark.asyncio
async def test_exec_setup_and_teardown(monkeypatch):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "exec",
    )
    p = UserProcess()

    await p.setup()

    assert p._process is None

    with nullcontext():
        await p.teardown()


@pytest.mark.parametrize("api_method", ["exec", "invoke"])
@pytest.mark.asyncio
async def test_teardown_without_setup(monkeypatch, api_method):
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        api_method,
    )
    process = UserProcess()

    with nullcontext():
        await process.teardown()


@pytest.mark.asyncio
async def test_invoke_call_timeout(local_s3, monkeypatch, capsys):
    async def _sleep(*_, **__):
        await asyncio.sleep(10)

    process = UserProcess()
    pk = str(uuid4())
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=f"tasks/{pk}",
        timeout=timedelta(seconds=1),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "invoke",
    )
    monkeypatch.setattr(process, "run_inference", _sleep)
    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 1
    assert int(result.invoke_duration.total_seconds()) == 1
    assert result.exec_duration is None  # should only be set for exec

    captured = capsys.readouterr()
    # "Time limit exceeded" must be the last log for the user error
    assert captured.err == (
        '{"log": "Time limit exceeded", "level": "ERROR", "source": "stderr", '
        f'"internal": false, "task": "{pk}"}}\n'
    )


@pytest.mark.asyncio
async def test_invoke_returns_0_on_201(mocker, monkeypatch):
    mocker.patch(
        "sagemaker_shim.models.httpx.AsyncClient.post",
        return_value=httpx.Response(201),
    )
    process = UserProcess()

    return_code = await process.invoke(timeout=timedelta(seconds=1))

    assert return_code == 0


@pytest.mark.parametrize(
    "client_kwargs, expectation",
    [
        ({"return_value": httpx.Response(201)}, nullcontext()),
        (
            {"side_effect": httpx.TimeoutException("error")},
            pytest.raises(UserSafeError, match="Invoke time limit exceeded"),
        ),
        (
            {"side_effect": httpx.ConnectError("error")},
            pytest.raises(
                UserSafeError, match="Could not connect to invoke endpoint"
            ),
        ),
        (
            {"side_effect": httpx.HTTPError("error")},
            pytest.raises(
                UserSafeError, match="HTTP error calling invoke endpoint"
            ),
        ),
        (
            {"return_value": httpx.Response(100)},
            pytest.raises(
                UserSafeError,
                match="Invoke endpoint returned status 100, expected 201",
            ),
        ),
        (
            {"return_value": httpx.Response(200)},
            pytest.raises(
                UserSafeError,
                match="Invoke endpoint returned status 200, expected 201",
            ),
        ),
        (
            {"return_value": httpx.Response(300)},
            pytest.raises(
                UserSafeError,
                match="Invoke endpoint returned status 300, expected 201",
            ),
        ),
        (
            {"return_value": httpx.Response(400)},
            pytest.raises(
                UserSafeError,
                match="Invoke endpoint returned status 400, expected 201",
            ),
        ),
        (
            {"return_value": httpx.Response(500)},
            pytest.raises(
                UserSafeError,
                match="Invoke endpoint returned status 500, expected 201",
            ),
        ),
    ],
)
@pytest.mark.asyncio
async def test_invoke_raises_on_non_201(
    mocker, monkeypatch, client_kwargs, expectation
):
    mocker.patch(
        "sagemaker_shim.models.httpx.AsyncClient.post",
        **client_kwargs,
    )
    process = UserProcess()

    with expectation:
        await process.invoke(timeout=timedelta(seconds=1))


@pytest.mark.asyncio
async def test_invoke_result_duration(local_s3, mocker, monkeypatch):
    mocker.patch(
        "sagemaker_shim.models.httpx.AsyncClient.post",
        return_value=httpx.Response(201),
    )
    process = UserProcess()
    pk = str(uuid4())
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=f"tasks/{pk}",
        timeout=timedelta(seconds=1),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "invoke",
    )
    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 0
    assert result.exec_duration is None
    assert result.invoke_duration is not None


@pytest.mark.asyncio
async def test_exec_result_duration(local_s3, monkeypatch):
    process = UserProcess()
    pk = str(uuid4())
    task = InferenceTask(
        pk=pk,
        inputs=[],
        output_bucket_name=local_s3.output_bucket_name,
        output_prefix=f"tasks/{pk}",
        timeout=timedelta(seconds=1),
    )
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_CMD_B64J",
        encode_b64j(val=["echo", "hello"]),
    )
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_USE_LINKED_INPUT", "False")
    monkeypatch.setenv("GRAND_CHALLENGE_COMPONENT_SET_EXTRA_GROUPS", "False")
    monkeypatch.setenv(
        "GRAND_CHALLENGE_COMPONENT_API_METHOD",
        "exec",
    )
    logging.config.dictConfig(LOGGING_CONFIG)

    async with get_s3_resources() as s3_resources:
        result = await task.run_inference(
            user_process=process, s3_resources=s3_resources
        )

    assert result.return_code == 0
    assert result.exec_duration is not None
    assert result.invoke_duration is None
